# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base form classes
"""

import json
import logging

import colander
import deform
from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttaweb.util import get_form_data, get_model_fields


log = logging.getLogger(__name__)


class FieldList(list):
    """
    Convenience wrapper for a form's field list.  This is a subclass
    of :class:`python:list`.

    You normally would not need to instantiate this yourself, but it
    is used under the hood for :attr:`Form.fields` as well as
    :attr:`~wuttaweb.grids.base.Grid.columns`.
    """

    def insert_before(self, field, newfield):
        """
        Insert a new field, before an existing field.

        :param field: String name for the existing field.

        :param newfield: String name for the new field, to be inserted
           just before the existing ``field``.
        """
        if field in self:
            i = self.index(field)
            self.insert(i, newfield)
        else:
            log.warning("field '%s' not found, will append new field: %s",
                        field, newfield)
            self.append(newfield)

    def insert_after(self, field, newfield):
        """
        Insert a new field, after an existing field.

        :param field: String name for the existing field.

        :param newfield: String name for the new field, to be inserted
           just after the existing ``field``.
        """
        if field in self:
            i = self.index(field)
            self.insert(i + 1, newfield)
        else:
            log.warning("field '%s' not found, will append new field: %s",
                        field, newfield)
            self.append(newfield)


class Form:
    """
    Base class for all forms.

    :param request: Reference to current :term:`request` object.

    :param fields: List of field names for the form.  This is
       optional; if not specified an attempt will be made to deduce
       the list automatically.  See also :attr:`fields`.

    :param schema: Colander-based schema object for the form.  This is
       optional; if not specified an attempt will be made to construct
       one automatically.  See also :meth:`get_schema()`.

    :param labels: Optional dict of default field labels.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    Form instances contain the following attributes:

    .. attribute:: request

       Reference to current :term:`request` object.

    .. attribute:: fields

       :class:`FieldList` instance containing string field names for
       the form.  By default, fields will appear in the same order as
       they are in this list.

    .. attribute:: schema

       :class:`colander:colander.Schema` object for the form.  This is
       optional; if not specified an attempt will be made to construct
       one automatically.

       See also :meth:`get_schema()`.

    .. attribute:: model_class

       Model class for the form, if applicable.  When set, this is
       usually a SQLAlchemy mapped class.  This (or
       :attr:`model_instance`) may be used instead of specifying the
       :attr:`schema`.

    .. attribute:: model_instance

       Optional instance from which initial form data should be
       obtained.  In simple cases this might be a dict, or maybe an
       instance of :attr:`model_class`.

       Note that this also may be used instead of specifying the
       :attr:`schema`, if the instance belongs to a class which is
       SQLAlchemy-mapped.  (In that case :attr:`model_class` can be
       determined automatically.)

    .. attribute:: readonly

       Boolean indicating the form does not allow submit.  In practice
       this means there will not even be a ``<form>`` tag involved.

       Default for this is ``False`` in which case the ``<form>`` tag
       will exist and submit is allowed.

    .. attribute:: readonly_fields

       A :class:`~python:set` of field names which should be readonly.
       Each will still be rendered but with static value text and no
       widget.

       This is only applicable if :attr:`readonly` is ``False``.

       See also :meth:`set_readonly()` and :meth:`is_readonly()`.

    .. attribute:: required_fields

       A dict of "required" field flags.  Keys are field names, and
       values are boolean flags indicating whether the field is
       required.

       Depending on :attr:`schema`, some fields may be "(not)
       required" by default.  However ``required_fields`` keeps track
       of any "overrides" per field.

       See also :meth:`set_required()` and :meth:`is_required()`.

    .. attribute:: action_url

       String URL to which the form should be submitted, if applicable.

    .. attribute:: cancel_url

       String URL to which the Cancel button should "always" redirect,
       if applicable.

       Code should not access this directly, but instead call
       :meth:`get_cancel_url()`.

    .. attribute:: cancel_url_fallback

       String URL to which the Cancel button should redirect, if
       referrer cannot be determined from request.

       Code should not access this directly, but instead call
       :meth:`get_cancel_url()`.

    .. attribute:: vue_tagname

       String name for Vue component tag.  By default this is
       ``'wutta-form'``.  See also :meth:`render_vue_tag()`.

    .. attribute:: align_buttons_right

       Flag indicating whether the buttons (submit, cancel etc.)
       should be aligned to the right of the area below the form.  If
       not set, the buttons are left-aligned.

    .. attribute:: auto_disable_submit

       Flag indicating whether the submit button should be
       auto-disabled, whenever the form is submitted.

    .. attribute:: button_label_submit

       String label for the form submit button.  Default is ``"Save"``.

    .. attribute:: button_icon_submit

       String icon name for the form submit button.  Default is ``'save'``.

    .. attribute:: button_type_submit

       Buefy type for the submit button.  Default is ``'is-primary'``,
       so for example:

       .. code-block:: html

          <b-button type="is-primary"
                    native-type="submit">
            Save
          </b-button>

       See also the `Buefy docs
       <https://buefy.org/documentation/button/#api-view>`_.

    .. attribute:: show_button_reset

       Flag indicating whether a Reset button should be shown.
       Default is ``False``.

    .. attribute:: show_button_cancel

       Flag indicating whether a Cancel button should be shown.
       Default is ``True``.

    .. attribute:: button_label_cancel

       String label for the form cancel button.  Default is
       ``"Cancel"``.

    .. attribute:: auto_disable_cancel

       Flag indicating whether the cancel button should be
       auto-disabled, whenever the button is clicked.  Default is
       ``True``.

    .. attribute:: validated

       If the :meth:`validate()` method was called, and it succeeded,
       this will be set to the validated data dict.

       Note that in all other cases, this attribute may not exist.
    """

    def __init__(
            self,
            request,
            fields=None,
            schema=None,
            model_class=None,
            model_instance=None,
            readonly=False,
            readonly_fields=[],
            required_fields={},
            labels={},
            action_url=None,
            cancel_url=None,
            cancel_url_fallback=None,
            vue_tagname='wutta-form',
            align_buttons_right=False,
            auto_disable_submit=True,
            button_label_submit="Save",
            button_icon_submit='save',
            button_type_submit='is-primary',
            show_button_reset=False,
            show_button_cancel=True,
            button_label_cancel="Cancel",
            auto_disable_cancel=True,
    ):
        self.request = request
        self.schema = schema
        self.readonly = readonly
        self.readonly_fields = set(readonly_fields or [])
        self.required_fields = required_fields or {}
        self.labels = labels or {}
        self.action_url = action_url
        self.cancel_url = cancel_url
        self.cancel_url_fallback = cancel_url_fallback
        self.vue_tagname = vue_tagname
        self.align_buttons_right = align_buttons_right
        self.auto_disable_submit = auto_disable_submit
        self.button_label_submit = button_label_submit
        self.button_icon_submit = button_icon_submit
        self.button_type_submit = button_type_submit
        self.show_button_reset = show_button_reset
        self.show_button_cancel = show_button_cancel
        self.button_label_cancel = button_label_cancel
        self.auto_disable_cancel = auto_disable_cancel

        self.config = self.request.wutta_config
        self.app = self.config.get_app()

        self.model_class = model_class
        self.model_instance = model_instance
        if self.model_instance and not self.model_class:
            self.model_class = type(self.model_instance)

        fields = fields or self.get_fields()
        if fields:
            self.set_fields(fields)
        else:
            self.fields = None

    def __contains__(self, name):
        """
        Custom logic for the ``in`` operator, to allow easily checking
        if the form contains a given field::

           myform = Form()
           if 'somefield' in myform:
               print("my form has some field")
        """
        return bool(self.fields and name in self.fields)

    def __iter__(self):
        """
        Custom logic to allow iterating over form field names::

           myform = Form(fields=['foo', 'bar'])
           for fieldname in myform:
               print(fieldname)
        """
        return iter(self.fields)

    @property
    def vue_component(self):
        """
        String name for the Vue component, e.g. ``'WuttaForm'``.

        This is a generated value based on :attr:`vue_tagname`.
        """
        words = self.vue_tagname.split('-')
        return ''.join([word.capitalize() for word in words])

    def get_cancel_url(self):
        """
        Returns the URL for the Cancel button.

        If :attr:`cancel_url` is set, its value is returned.

        Or, if the referrer can be deduced from the request, that is
        returned.

        Or, if :attr:`cancel_url_fallback` is set, that value is
        returned.

        As a last resort the "default" URL from
        :func:`~wuttaweb.subscribers.request.get_referrer()` is
        returned.
        """
        # use "permanent" URL if set
        if self.cancel_url:
            return self.cancel_url

        # nb. use fake default to avoid normal default logic;
        # that way if we get something it's a real referrer
        url = self.request.get_referrer(default='NOPE')
        if url and url != 'NOPE':
            return url

        # use fallback URL if set
        if self.cancel_url_fallback:
            return self.cancel_url_fallback

        # okay, home page then (or whatever is the default URL)
        return self.request.get_referrer()

    def set_fields(self, fields):
        """
        Explicitly set the list of form fields.

        This will overwrite :attr:`fields` with a new
        :class:`FieldList` instance.

        :param fields: List of string field names.
        """
        self.fields = FieldList(fields)

    def set_readonly(self, key, readonly=True):
        """
        Enable or disable the "readonly" flag for a given field.

        When a field is marked readonly, it will be shown in the form
        but there will be no editable widget.  The field is skipped
        over (not saved) when form is submitted.

        See also :meth:`is_readonly()`; this is tracked via
        :attr:`readonly_fields`.

        :param key: String key (fieldname) for the field.

        :param readonly: New readonly flag for the field.
        """
        if readonly:
            self.readonly_fields.add(key)
        else:
            if key in self.readonly_fields:
                self.readonly_fields.remove(key)

    def is_readonly(self, key):
        """
        Returns boolean indicating if the given field is marked as
        readonly.

        See also :meth:`set_readonly()`.

        :param key: Field key/name as string.
        """
        if self.readonly_fields:
            if key in self.readonly_fields:
                return True
        return False

    def set_required(self, key, required=True):
        """
        Enable or disable the "required" flag for a given field.

        When a field is marked required, a value must be provided
        or else it fails validation.

        In practice if a field is "not required" then a default
        "empty" value is assumed, should the user not provide one.

        See also :meth:`is_required()`; this is tracked via
        :attr:`required_fields`.

        :param key: String key (fieldname) for the field.

        :param required: New required flag for the field.  Usually a
           boolean, but may also be ``None`` to remove any flag and
           revert to default behavior for the field.
        """
        self.required_fields[key] = required

    def is_required(self, key):
        """
        Returns boolean indicating if the given field is marked as
        required.

        See also :meth:`set_required()`.

        :param key: Field key/name as string.

        :returns: Value for the flag from :attr:`required_fields` if
           present; otherwise ``None``.
        """
        return self.required_fields.get(key, None)

    def set_label(self, key, label):
        """
        Set the label for given field name.

        See also :meth:`get_label()`.
        """
        self.labels[key] = label

        # update schema if necessary
        if self.schema and key in self.schema:
            self.schema[key].title = label

    def get_label(self, key):
        """
        Get the label for given field name.

        Note that this will always return a string, auto-generating
        the label if needed.

        See also :meth:`set_label()`.
        """
        return self.labels.get(key, self.app.make_title(key))

    def get_fields(self):
        """
        Returns the official list of field names for the form, or
        ``None``.

        If :attr:`fields` is set and non-empty, it is returned.

        Or, if :attr:`schema` is set, the field list is derived
        from that.

        Or, if :attr:`model_class` is set, the field list is derived
        from that, via :meth:`get_model_fields()`.

        Otherwise ``None`` is returned.
        """
        if hasattr(self, 'fields') and self.fields:
            return self.fields

        if self.schema:
            return [field.name for field in self.schema]

        fields = self.get_model_fields()
        if fields:
            return fields

    def get_model_fields(self, model_class=None):
        """
        This method is a shortcut which calls
        :func:`~wuttaweb.util.get_model_fields()`.

        :param model_class: Optional model class for which to return
           fields.  If not set, the form's :attr:`model_class` is
           assumed.
        """
        return get_model_fields(self.config,
                                model_class=model_class or self.model_class)

    def get_schema(self):
        """
        Return the :class:`colander:colander.Schema` object for the
        form, generating it automatically if necessary.

        Note that if :attr:`schema` is already set, that will be
        returned as-is.
        """
        if not self.schema:

            # get fields
            fields = self.get_fields()
            if not fields:
                raise NotImplementedError

            # make basic schema
            schema = colander.Schema()
            for name in fields:
                schema.add(colander.SchemaNode(
                    colander.String(),
                    name=name))

            # apply required flags
            for key, required in self.required_fields.items():
                if key in schema:
                    if required is False:
                        # TODO: (why) should we not use colander.null here?
                        #schema[key].missing = colander.null
                        schema[key].missing = None

            self.schema = schema

        return self.schema

    def get_deform(self):
        """
        Return the :class:`deform:deform.Form` instance for the form,
        generating it automatically if necessary.
        """
        if not hasattr(self, 'deform_form'):
            model = self.app.model
            schema = self.get_schema()
            kwargs = {}

            if self.model_instance:
                if isinstance(self.model_instance, model.Base):
                    kwargs['appstruct'] = dict(self.model_instance)
                else:
                    kwargs['appstruct'] = self.model_instance

                # TODO: ugh why is this necessary?
                for key, value in list(kwargs['appstruct'].items()):
                    if value is None:
                        kwargs['appstruct'][key] = colander.null

            form = deform.Form(schema, **kwargs)
            self.deform_form = form

        return self.deform_form

    def render_vue_tag(self, **kwargs):
        """
        Render the Vue component tag for the form.

        By default this simply returns:

        .. code-block:: html

           <wutta-form></wutta-form>

        The actual output will depend on various form attributes, in
        particular :attr:`vue_tagname`.
        """
        return HTML.tag(self.vue_tagname, **kwargs)

    def render_vue_template(
            self,
            template='/forms/vue_template.mako',
            **context):
        """
        Render the Vue template block for the form.

        This returns something like:

        .. code-block:: none

           <script type="text/x-template" id="wutta-form-template">
             <form>
               <!-- fields etc. -->
             </form>
           </script>

        .. todo::

           Why can't Sphinx render the above code block as 'html' ?

           It acts like it can't handle a ``<script>`` tag at all?

        Actual output will of course depend on form attributes, i.e.
        :attr:`vue_tagname` and :attr:`fields` list etc.

        :param template: Path to Mako template which is used to render
           the output.
        """
        context['form'] = self
        context['dform'] = self.get_deform()
        context.setdefault('form_attrs', {})
        context.setdefault('request', self.request)

        # auto disable button on submit
        if self.auto_disable_submit:
            context['form_attrs']['@submit'] = 'formSubmitting = true'

        output = render(template, context)
        return HTML.literal(output)

    def render_vue_field(self, fieldname, readonly=None):
        """
        Render the given field completely, i.e. ``<b-field>`` wrapper
        with label and containing a widget.

        Actual output will depend on the field attributes etc.
        Typical output might look like:

        .. code-block:: html

           <b-field label="Foo"
                    horizontal
                    type="is-danger"
                    message="something went wrong!">
              <!-- widget element(s) -->
           </b-field>
        """
        # readonly comes from: caller, field flag, or form flag
        if readonly is None:
            readonly = self.is_readonly(fieldname)
            if not readonly:
                readonly = self.readonly

        # but also, fields not in deform/schema must be readonly
        dform = self.get_deform()
        if not readonly and fieldname not in dform:
            readonly = True

        # render the field widget or whatever
        if fieldname in dform:

            # render proper widget if field is in deform/schema
            field = dform[fieldname]
            kw = {}
            if readonly:
                kw['readonly'] = True
            html = field.serialize(**kw)

        else:
            # render static text if field not in deform/schema
            # TODO: need to abstract this somehow
            if self.model_instance:
                html = str(self.model_instance[fieldname])
            else:
                html = ''

        # mark all that as safe
        html = HTML.literal(html)

        # render field label
        label = self.get_label(fieldname)

        # b-field attrs
        attrs = {
            ':horizontal': 'true',
            'label': label,
        }

        # next we will build array of messages to display..some
        # fields always show a "helptext" msg, and some may have
        # validation errors..
        field_type = None
        messages = []

        # show errors if present
        errors = self.get_field_errors(fieldname)
        if errors:
            field_type = 'is-danger'
            messages.extend(errors)

        # ..okay now we can declare the field messages and type
        if field_type:
            attrs['type'] = field_type
        if messages:
            if len(messages) == 1:
                msg = messages[0]
                if msg.startswith('`') and msg.endswith('`'):
                    attrs[':message'] = msg
                else:
                    attrs['message'] = msg
            # TODO
            # else:
            #     # nb. must pass an array as JSON string
            #     attrs[':message'] = '[{}]'.format(', '.join([
            #         "'{}'".format(msg.replace("'", r"\'"))
            #         for msg in messages]))

        return HTML.tag('b-field', c=[html], **attrs)

    def get_field_errors(self, field):
        """
        Return a list of error messages for the given field.

        Not useful unless a call to :meth:`validate()` failed.
        """
        dform = self.get_deform()
        if field in dform:
            error = dform[field].errormsg
            if error:
                return [error]
        return []

    def get_vue_field_value(self, field):
        """
        This method returns a JSON string which will be assigned as
        the initial model value for the given field.  This JSON will
        be written as part of the overall response, to be interpreted
        on the client side.

        Again, this must return a *string* such as:

        * ``'null'``
        * ``'{"foo": "bar"}'``

        In practice this calls :meth:`jsonify_value()` to convert the
        ``field.cstruct`` value to string.
        """
        if isinstance(field, str):
            dform = self.get_deform()
            field = dform[field]

        return self.jsonify_value(field.cstruct)

    def jsonify_value(self, value):
        """
        Convert a Python value to JSON string.

        See also :meth:`get_vue_field_value()`.
        """
        if value is colander.null:
            return 'null'

        return json.dumps(value)

    def validate(self):
        """
        Try to validate the form, using data from the :attr:`request`.

        Uses :func:`~wuttaweb.util.get_form_data()` to retrieve the
        form data from POST or JSON body.

        If the form data is valid, the data dict is returned.  This
        data dict is also made available on the form object via the
        :attr:`validated` attribute.

        However if the data is not valid, ``False`` is returned, and
        there will be no :attr:`validated` attribute.  In that case
        you should inspect the form errors to learn/display what went
        wrong for the user's sake.  See also
        :meth:`get_field_errors()`.

        This uses :meth:`deform:deform.Field.validate()` under the
        hood.

        .. warning::

           Calling ``validate()`` on some forms will cause the
           underlying Deform and Colander structures to mutate.  In
           particular, all :attr:`readonly_fields` will be *removed*
           from the :attr:`schema` to ensure they are not involved in
           the validation.

        :returns: Data dict, or ``False``.
        """
        if hasattr(self, 'validated'):
            del self.validated

        if self.request.method != 'POST':
            return False

        # remove all readonly fields from deform / schema
        dform = self.get_deform()
        if self.readonly_fields:
            schema = self.get_schema()
            for field in self.readonly_fields:
                if field in schema:
                    del schema[field]
                    dform.children.remove(dform[field])

        # let deform do real validation
        controls = get_form_data(self.request).items()
        try:
            self.validated = dform.validate(controls)
        except deform.ValidationFailure:
            return False

        return self.validated
